/* $Id: bac_load.c,v 1.72 2002/12/20 12:38:12 oohara Exp $ */

#include <stdio.h>
/* qsort */
#include <stdlib.h>
/* memset */
#include <string.h>

/* libdb3 */
#include <db.h>

#include "bac_array.h"
#include "bac_entry.h"
#include "bac_string_to_array.h"

#include "bac_load.h"

static bac_array *bac_load_add_new_entry(bac_array *array, const char *string,
                                         int n, int flag, DBC *cursorp);

/* return a new bac_array which consists of loaded entries
 * no_entry (arg 6) is set to 1 if bac_load() returns NULL because
 * there is no entry, 0 if not
 * if an error happens, load the entries that are loaded
 * before the error
 * assumes all data in the database are null-terminated
 * also assumes there is at least one entry
 * note that the returned array may have more than n entries in it
 */
bac_array *
bac_load(const char *dir, const char *file, int n, int flag, int mode,
         int *no_entry)
{
  DB *dbp = NULL;
  DB_ENV *dbenvp = NULL;
  DBC *cursorp = NULL;
  DBT key;
  DBT data;
  u_int32_t dbflag = 0;
  int status;
  bac_array *p = NULL;
  bac_array *p_temp = NULL;

  *no_entry = 0;

  /* sanity check */
  if (dir == NULL)
    return NULL;
  if (file == NULL)
    return NULL;
  if (n <= 0)
    return NULL;

  /* this must be before db_create */
  if (db_env_create(&dbenvp, 0) != 0)
  {
    fprintf(stderr, "bac_load: db_env_create failed\n");
    return NULL;
  }
  /* this must be before DB->open */
  /* this directory must be present on the system before
   * this program is called (DB_CREATE creates only files in that
   * directory, not the directory itself)
   */
  /* DB_INIT_CDB and DB_INIT_MPOOL means Concurrent Data Store */
  if (dbenvp->open(dbenvp, dir,
                   DB_INIT_CDB | DB_INIT_MPOOL | DB_CREATE, mode) != 0)
  {
    fprintf(stderr, "bac_load: DBENV->open(%s) failed\n", dir);
    dbenvp->close(dbenvp, 0);
    return NULL;
  }
  if (db_create(&dbp, dbenvp, 0) != 0)
  {
    fprintf(stderr, "bac_load: db_create failed\n");
    dbenvp->close(dbenvp, 0);
    return NULL;
  }
  if (dbp->open(dbp, file, NULL, DB_RECNO,
                DB_CREATE, mode) != 0)
  {
    fprintf(stderr, "bac_load: DB->open(%s) failed\n", file);
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return NULL;
  }

  if (flag & BAC_LOAD_FLAG_NO_DELETE)
    dbflag = 0;
  else
    dbflag = DB_WRITECURSOR;
  if (dbp->cursor(dbp, NULL, &cursorp, dbflag) != 0)
  {
    fprintf(stderr, "bac_load: DB->cursor failed\n");
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return NULL;
  }
  /* for compatibility with later version of libdb */
  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));
  status = cursorp->c_get(cursorp, &key, &data, DB_FIRST);
  if (status == DB_NOTFOUND)
  {
    /* there is no element in the database */
    /* be quiet, there may be --dump option
     * the calling function should produce an appropriate error message
     */
    *no_entry = 1;
    cursorp->c_close(cursorp);
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    /* there is no special return value here --- after all,
     * you have no array */
    return NULL;
  }
  else if (status != 0)
  {
    fprintf(stderr, "bac_load: DBcursor->c_get(DB_FIRST) failed\n");
    cursorp->c_close(cursorp);
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return NULL;
  }
  if (((char *) (data.data))[(data.size / sizeof(char)) - 1] != '\0' )
  {
    fprintf(stderr, "bac_load: data.data is not null-terminated\n");
    cursorp->c_close(cursorp);
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return NULL;
  }

  p_temp = bac_load_add_new_entry(p, (char *) (data.data), n, flag, cursorp);
  if (p_temp == NULL)
  {
    fprintf(stderr, "bac_load: bac_load_add_new_entry failed\n");
    cursorp->c_close(cursorp);
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return NULL;
  }
  p = p_temp;

  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));
  while(1 == 1)
  {
    status = cursorp->c_get(cursorp, &key, &data, DB_NEXT);
    if (status == DB_NOTFOUND)
    {
      /* no more entry in the database */
      break;
    }
    else if (status != 0)
    {
      fprintf(stderr, "bac_load: DBcursor->c_get(DB_NEXT) failed\n");
      cursorp->c_close(cursorp);
      dbp->close(dbp, 0);
      dbenvp->close(dbenvp, 0);
      return p;
    }
    else
    {
      if (((char *) (data.data))[(data.size / sizeof(char)) - 1] != '\0' )
      {
        fprintf(stderr, "bac_load: data.data is not null-terminated\n");
        cursorp->c_close(cursorp);
        dbp->close(dbp, 0);
        dbenvp->close(dbenvp, 0);
        return p;
      }

      p_temp = bac_load_add_new_entry(p, (char *) (data.data), n, flag,
                                      cursorp);
      if (p_temp == NULL)
      {
        fprintf(stderr, "bac_load: bac_load_add_new_entry failed\n");
        cursorp->c_close(cursorp);
        dbp->close(dbp, 0);
        dbenvp->close(dbenvp, 0);
        return p;
      }
      p = p_temp;

      memset(&key, 0, sizeof(key));
      memset(&data, 0, sizeof(data));
    }
  }

  cursorp->c_close(cursorp);
  dbp->close(dbp, 0);
  dbenvp->close(dbenvp, 0);

  return p;
}

/* assumes array (arg 1) is sorted as long as it is not NULL
 */
static bac_array *
bac_load_add_new_entry(bac_array *array, const char *string, int n, int flag,
                       DBC *cursorp)
{
  bac_array *array_new = NULL;
  bac_array *array_temp = NULL;
  int in_nth;

  array_new = bac_string_to_array(string);
  if (array_new == NULL)
  {
    fprintf(stderr, "bac_load_add_new_entry: bac_string_to_array failed\n");
    return NULL;
  }
  /* simple !(flag & BAC_LOAD_FLAG_NO_SORT) is wrong */
  if (!((flag & BAC_LOAD_FLAG_NO_SORT) == BAC_LOAD_FLAG_NO_SORT))
    qsort((void *) array_new->p, array_new->n, sizeof(bac_entry *),
          (int (*)(const void *, const void *)) bac_compare_entry);

  if ((array != NULL) && (array_new != NULL) && (array->n >= n)
      && (bac_compare_entry((void *) (&(array->p[array->n - 1])),
                            (void *) (&(array_new->p[0]))) <= 0))
  {
    /* no entry in array_new is above nth */
    /* false */
    in_nth = 0;
  }
  else
  {
    /* true */
    in_nth = 1;
  }

  if ((!(flag & BAC_LOAD_FLAG_NO_DELETE)) && (!in_nth))
  {
    if (cursorp->c_del(cursorp, 0) != 0)
    {
      fprintf(stderr, "bac_load_add_new_entry: "
              "DBcursor->c_del failed\n");
      /* we may quit now because we return NULL regardless of flag */
      if (array_new != NULL)
        bac_array_delete(array_new);
      return NULL;
    }
  }

  if ((flag & BAC_LOAD_FLAG_LOAD_ALL) || (in_nth))
  {
    array_temp = bac_array_add_array(array, array_new);
    if (array_temp == NULL)
    {
      fprintf(stderr, "bac_load: bac_array_add_array failed\n");
      if (array_new != NULL)
        bac_array_delete(array_new);
      return NULL;
    }
    array = array_temp;
    if (!((flag & BAC_LOAD_FLAG_NO_SORT) == BAC_LOAD_FLAG_NO_SORT))
      qsort((void *) array->p, array->n, sizeof(bac_entry *),
            (int (*)(const void *, const void *)) bac_compare_entry);
  }

  return array;
}
