/* $Id: bac_truncate.c,v 1.27 2002/12/20 21:48:17 oohara Exp $ */

#include <stdio.h>
/* qsort */
#include <stdlib.h>
/* memset */
#include <string.h>

/* libdb3 */
#include <db.h>

#include "bac_array.h"
#include "bac_entry.h"
#include "bac_string_to_array.h"

#include "bac_truncate.h"

static int bac_truncate_one_step(bac_entry *entry, const char *string,
                                 DBC *cursorp);

/* delete all entries that are (strictly) below entry (arg 3)
 * in the database
 * return 0 on success, 1 on error
 * assumes there is at least one entry --- if not, it is an error
 */
int
bac_truncate(const char *dir, const char *file, bac_entry *entry, int mode)
{
  DB *dbp = NULL;
  DB_ENV *dbenvp = NULL;
  DBC *cursorp = NULL;
  DBT key;
  DBT data;
  int status;

  /* sanity check */
  if (dir == NULL)
    return 1;
  if (file == NULL)
    return 1;
  if (entry == NULL)
    return 1;

  /* this must be before db_create */
  if (db_env_create(&dbenvp, 0) != 0)
  {
    fprintf(stderr, "bac_truncate: db_env_create failed\n");
    return 1;
  }
  /* this must be before DB->open */
  /* this directory must be present on the system before
   * this program is called (DB_CREATE creates only files in that
   * directory, not the directory itself)
   */
  /* DB_INIT_CDB and DB_INIT_MPOOL means Concurrent Data Store */
  if (dbenvp->open(dbenvp, dir,
                   DB_INIT_CDB | DB_INIT_MPOOL | DB_CREATE, mode) != 0)
  {
    fprintf(stderr, "bac_truncate: DBENV->open failed\n");
    dbenvp->close(dbenvp, 0);
    return 1;
  }
  if (db_create(&dbp, dbenvp, 0) != 0)
  {
    fprintf(stderr, "bac_truncate: db_create failed\n");
    dbenvp->close(dbenvp, 0);
    return 1;
  }
  if (dbp->open(dbp, file, NULL, DB_RECNO,
                DB_CREATE, mode) != 0)
  {
    fprintf(stderr, "bac_truncate: DB->open failed\n");
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return 1;
  }
  if (dbp->cursor(dbp, NULL, &cursorp, DB_WRITECURSOR) != 0)
  {
    fprintf(stderr, "bac_truncate: DB->cursor failed\n");
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return 1;
  }

  /* for compatibility with later version of libdb */
  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));
  status = cursorp->c_get(cursorp, &key, &data, DB_FIRST);
  if (status == DB_NOTFOUND)
  {
    /* there is no element in the database, so nothing to do */
    cursorp->c_close(cursorp);
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return 0;
  }
  else if (status != 0)
  {
    fprintf(stderr, "bac_truncate: DBcursor->c_get(DB_FIRST) failed\n");
    cursorp->c_close(cursorp);
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return 1;
  }
  if (((char *) (data.data))[(data.size / sizeof(char)) - 1] != '\0' )
  {
    fprintf(stderr, "bac_truncate: data.data is not null-terminated\n");
    cursorp->c_close(cursorp);
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return 1;
  }

  if (bac_truncate_one_step(entry, (char *) (data.data), cursorp) != 0)
  {
    fprintf(stderr, "bac_truncate: bac_truncate_one_step failed\n");
    cursorp->c_close(cursorp);
    dbp->close(dbp, 0);
    dbenvp->close(dbenvp, 0);
    return 1;
  }
  memset(&key, 0, sizeof(key));
  memset(&data, 0, sizeof(data));

  while(1 == 1)
  {
    status = cursorp->c_get(cursorp, &key, &data, DB_NEXT);
    if (status == DB_NOTFOUND)
    {
      /* no more entry in the database */
      break;
    }
    else if (status != 0)
    {
      fprintf(stderr, "bac_truncate: DBcursor->c_get(DB_NEXT) failed\n");
      cursorp->c_close(cursorp);
      dbp->close(dbp, 0);
      dbenvp->close(dbenvp, 0);
      return 1;
    }
    else
    {
      if (((char *) (data.data))[(data.size / sizeof(char)) - 1] != '\0' )
      {
        fprintf(stderr, "bac_truncate: data.data is not null-terminated\n");
        cursorp->c_close(cursorp);
        dbp->close(dbp, 0);
        dbenvp->close(dbenvp, 0);
        return 1;
      }
      if (bac_truncate_one_step(entry, (char *) (data.data), cursorp) != 0)
      {
        fprintf(stderr, "bac_truncate: bac_truncate_one_step failed\n");
        cursorp->c_close(cursorp);
        dbp->close(dbp, 0);
        dbenvp->close(dbenvp, 0);
        return 1;
      }
      memset(&key, 0, sizeof(key));
      memset(&data, 0, sizeof(data));
    }
  }

  cursorp->c_close(cursorp);
  dbp->close(dbp, 0);
  dbenvp->close(dbenvp, 0);

  return 0;
}

/* return 0 on success, 1 on error */
static int
bac_truncate_one_step(bac_entry *entry, const char *string, DBC *cursorp)
{
  bac_array *temp;
  int compare_result;

  temp = bac_string_to_array(string);
  if (temp == NULL)
  {
    fprintf(stderr, "bac_truncate_one_step: bac_string_to_array failed\n");
    return 1;
  }
  qsort((void *) temp->p, temp->n, sizeof(bac_entry *),
        (int (*)(const void *, const void *)) bac_compare_entry);

  compare_result = bac_compare_entry((void *) (&(temp->p[0])),
                                     (void *) (&entry));
  if (compare_result > 0)
  {
    if (cursorp->c_del(cursorp, 0) != 0)
    {
      fprintf(stderr, "bac_truncate: DBcursor->c_del failed\n");
      if (temp != NULL)
        bac_array_delete(temp);
      return 1;
    }
  }
  if (temp != NULL)
    bac_array_delete(temp);

  return 0;
}
